//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Sun May 13 14:15:43 PDT 2001
// Last Modified: Sun May 13 14:15:46 PDT 2001
// Filename:      ...sig/examples/all/kskey.cpp
// Web Address:   http://sig.sapp.org/examples/museinfo/humdrum/kskey.cpp
// Syntax:        C++; museinfo
//
// Description:   Does the Krumhansl-Schmuckler key-finding algorithm.
// 

#include "humdrum.h"

// function declarations
void   checkOptions             (Options& opts, int argc, char* argv[]);
void   example                  (void);
void   printAnalysis            (int bestkey, Array<double>& scores,
                                 Array<double>& durhist);
void   usage                    (const char* command);

// user interface variables
Options      options;            // database for command-line arguments
int          frequencyQ  = 0;    // used with the -f option
int          allQ  = 0;          // used with the -a option
int          rhythmQ = 1;        // used with the -q option
int          binaryQ = 0;        // used with the -b option

///////////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[]) {
   HumdrumFile infile;

   // process the command-line options
   checkOptions(options, argc, argv);

   // figure out the number of input files to process
   int numinputs = options.getArgCount();

   Array<double> absbeat;
   Array<int>    pitch;
   Array<double> duration;
   Array<double> level;
   Array<double> distribution(12);
   Array<double> scores(24);
  
   int bestkey = 0;
   int i, j;
   for (i=0; i<numinputs || i==0; i++) {
      infile.clear();

      // if no command-line arguments read data file from standard input
      if (numinputs < 1) {
         infile.read(cin);
      } else {
         infile.read(options.getArg(i+1));
      }

      infile.getNoteArray(absbeat, pitch, duration, level);
      for (j=0; j<pitch.getSize(); j++) {
         pitch[j] = Convert::base40ToMidiNoteNumber(pitch[j]);
      }
      bestkey = analyzeKeyKS(scores.getBase(), distribution.getBase(),
            pitch.getBase(), duration.getBase(), pitch.getSize(), rhythmQ,
                  binaryQ);
      printAnalysis(bestkey, scores, distribution);
   }

   return 0;
}


///////////////////////////////////////////////////////////////////////////

//////////////////////////////
//
// printAnalysis -- 
//

void printAnalysis(int bestkey, Array<double>& scores, Array<double>& durhist) {
   char buffer[64] = {0};
   if (bestkey < 12) {
      cout << "The best key is: " 
           << Convert::base12ToKern(buffer, bestkey+12*3)
           << " Major." << "\n";
   } else {
      cout << "The best key is: " 
           << Convert::base12ToKern(buffer, bestkey+12*2)
           << " Minor." << "\n";
   }
   int i;

   if (allQ) {
      for (i=0; i<12; i++) {
         cout << "Major[" << i << "] = " << scores[i] << "\t\t\t"
              << "Minor[" << i << "] = " << scores[i+12] << "\n";
      }
   }
   if (frequencyQ) {
      for (i=0; i<12; i++) {
         cout << "Pitch[" << i << "] = " << durhist[i] << "\n";
      }
   }
}



//////////////////////////////
//
// checkOptions -- validate and process command-line options.
//

void checkOptions(Options& opts, int argc, char* argv[]) {
   opts.define("a|all=b",             "show all scores");   
   opts.define("q|quarter=b",         "treat all rhytms as quarter notes");   
   opts.define("b|binary=b",          "do binary correlation");   
   opts.define("f|frequency|freq=b",  "show pitch frequencies");   

   opts.define("debug=b",       "trace input parsing");   
   opts.define("author=b",      "author of the program");   
   opts.define("version=b",     "compilation information"); 
   opts.define("example=b",     "example usage"); 
   opts.define("h|help=b",      "short description"); 
   opts.process(argc, argv);
   
   // handle basic options:
   if (opts.getBoolean("author")) {
      cout << "Written by Craig Stuart Sapp, "
           << "craig@ccrma.stanford.edu, May 2001" << endl;
      exit(0);
   } else if (opts.getBoolean("version")) {
      cout << argv[0] << ", version: 6 May 2002" << endl;
      cout << "compiled: " << __DATE__ << endl;
      cout << MUSEINFO_VERSION << endl;
      exit(0);
   } else if (opts.getBoolean("help")) {
      usage(opts.getCommand());
      exit(0);
   } else if (opts.getBoolean("example")) {
      example();
      exit(0);
   }

   allQ = opts.getBoolean("all");
   rhythmQ = !opts.getBoolean("quarter");
   binaryQ = opts.getBoolean("binary");
   frequencyQ = opts.getBoolean("frequency");
}



//////////////////////////////
//
// example -- example usage of the maxent program
//

void example(void) {
   cout <<
   "                                                                        \n"
   << endl;
}



//////////////////////////////
//
// usage -- gives the usage statement for the quality program
//

void usage(const char* command) {
   cout <<
   "                                                                        \n"
   << endl;
}


